import math
import pygame

from designer.colors import _process_color
from designer.helpers import get_width, get_height
from designer.objects.designer_object import DesignerObject
from designer.core.internal_image import InternalImage, DesignerSurface
from designer.utilities.vector import Vec2D
from designer.utilities.util import _anchor_offset

class Ellipse(DesignerObject):
    DEFAULT_BORDER_WIDTH = 1
    FIELDS = (*DesignerObject.FIELDS, 'color', 'border')

    def __init__(self, center, width, height, anchor, color, border):
        '''
        Creates an ellipse Designer Object on the window.

        :param left: x coordinate of top left corner of ellipse
        :type left: int
        :param top: y coordinate of top left corner of ellipse
        :type top: int
        :param width: width of ellipse to be drawn
        :type width: int
        :param height: height of ellipse to be drawn
        :type height: int
        :param anchor: the anchor to draw the circle at
        :type anchor: str
        :param color: color of ellipse
        :type color: str or List[str]
        :param border: the width of the circle's line (0 is used for a filled circle)
        :type border: int
        '''
        super().__init__()

        x, y = center
        x = x if x is not None else get_width() / 2
        y = y if y is not None else get_height() / 2
        center = x, y

        self._pos = center
        self._anchor = anchor
        # Ellipse specific data
        self._size = Vec2D(width, height)
        self._color = color
        self._border = border

        # And draw!
        self._redraw_internal_image()

    def _recalculate_offset(self):
        size = self._size * self._scale
        offset = _anchor_offset(self._anchor, size[0], size[1])
        self._offset = Vec2D(offset) - self._transform_offset

    def _redraw_internal_image(self):
        # No flipping
        # Scaling
        width = self._size[0] * self._scale[0]
        height = self._size[1] * self._scale[1]
        size = (int(width), int(height))
        color = _process_color(self._color)
        if size[0] <= 0 or size[1] <= 0:
            target = InternalImage(size=(1, 1))
            target.fill(color)
            target = self._transform_image._surf
            self._recalculate_offset()
            self._expire_static()
            return
        new_image = InternalImage(size=size)
        new_image.draw_ellipse(color, (0, 0), size, self._border or 0)
        # Rotation
        if self._angle != 0:
            angle = self._angle % 360
            old = Vec2D(new_image.rect.center)
            source = pygame.transform.rotate(new_image._surf, angle).convert_alpha()
            new = source.rect.center
            self._transform_offset = old - new
        self._transform_image = new_image._surf
        self._recalculate_offset()
        self._expire_static()

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value
        self._redraw_internal_image()

    @property
    def border(self):
        return self._border

    @border.setter
    def border(self, value):
        self._border = value
        self._redraw_internal_image()


def ellipse(color, width=None, height=None, x=None, y=None, anchor='center', border=None):
    '''
    Function to make an ellipse.

    :param color: color of ellipse
    :type color: str or List[str]
    :return: Ellipse object created
    '''
    if x is not None and y is None:
        x, y = x
    elif x is None and y is None and height is not None:
        if not isinstance(height, (int, float)):
            x, y = height
            width, height = width
    elif height is None:
        width, height = width
    return Ellipse((x, y), width, height, anchor, color, border)



import pygame
import math

from designer.colors import _process_color
from designer.helpers import get_width, get_height
from designer.objects.designer_object import DesignerObject
from designer.core.internal_image import InternalImage, DesignerSurface
from designer.utilities.vector import Vec2D
from designer.utilities.util import _anchor_offset

class Text(DesignerObject):
    DEFAULT_FONT_SIZE = 16
    DEFAULT_FONT_COLOR = 'black'
    DEFAULT_FONT_NAME = 'Arial'
    FONTS = {}
    FIELDS = (*DesignerObject.FIELDS, 'text', 'color', 'font', 'text_size')

    def __init__(self, center, anchor, text_string, text_size, color, font):
        """
        Creates Text Designer Object on window

        :param center: x, y coordinates of center of circle
        :type center: Tuple[int]
        :param anchor: the anchor to draw the circle at
        :type anchor: str
        :param color: color of text
        :type color: str or List[str]
        :param text: text to be written on window
        :type text: str
        :param text_size: font size of text
        :type text_size: int
        """
        super().__init__()

        x, y = center
        x = x if x is not None else get_width() / 2
        y = y if y is not None else get_height() / 2
        center = x, y

        self._pos = center
        self._anchor = anchor
        # Text Specific data
        self._text = text_string
        self._text_size = text_size
        self._color = color
        self._font_name = font
        self._update_font()

        # Draw the actual circle image
        self._redraw_internal_image()

    def _recalculate_offset(self):
        if self._transform_image is None:
            return
        size = self._scale * self._transform_image.get_size()
        offset = _anchor_offset(self._anchor, size[0], size[1])
        self._offset = Vec2D(offset) - self._transform_offset

    def _redraw_internal_image(self):
        text_surface = self._font.render(str(self.text), True, _process_color(self.color))
        target = InternalImage(size=text_surface.get_size())
        target._surf.blit(text_surface, (0, 0))
        self._default_redraw_transforms(target)

    @classmethod
    def _get_font(cls, font, text_size):
        if (font, text_size) not in cls.FONTS:
            cls.FONTS[(font, text_size)] = pygame.font.SysFont(font, text_size)
        return cls.FONTS[(font, text_size)]

    def _update_font(self):
        self._font = self._get_font(self._font_name, self._text_size)

    @property
    def text_size(self):
        return self._text_size

    @text_size.setter
    def text_size(self, value):
        self._text_size = value
        self._update_font()
        self._redraw_internal_image()

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value
        self._redraw_internal_image()

    @property
    def font(self):
        return self._font

    @font.setter
    def font(self, value):
        self._font_name = value
        self._update_font()
        self._redraw_internal_image()

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        self._text = value
        self._redraw_internal_image()


def text(color, text, text_size=Text.DEFAULT_FONT_SIZE,
         x=None, y=None, anchor='center', font_name=Text.DEFAULT_FONT_NAME):
    '''
       Function to create text.

       :param text: text to appear on window
       :type text: str
       :param color: color of text
       :type color: str or List[str]
       :return: Text object created
       '''
    if x is not None and y is None:
        try:
            x, y = x
        except TypeError as e:
            pass
    return Text((x, y), anchor, text, text_size, color, font_name)


"""This module contains classes to handle memoization, a time-saving method that
caches previously seen results from function calls."""


class Memoize:
    """
    This is a decorator to allow memoization of function calls. It is a
    completely dumb cache, and will cache anything given to it indefinitely.

    :param object func: Any function (although any object will work).
    .. warning:: This may be deprecated.
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        """
        Attempts to return the results of this function call from the cache;
        if it can't find it, then it will execute the function and add it to the
        cache.
        """
        try:
            return self.cache[args]
        except KeyError:
            res = self.func(*args)
            self.cache[args] = res
            return res
        except TypeError:
            print("WARNING: Unhashable type passed to memoize."
                  "Reconsider using this decorator.")
            return self.func(*args)


class SmartMemoize:
    """
    This is a decorator to allow memoization of function calls. Its cache
    is cleared on scene changes, and also clears items from the cache which
    haven't been used in at least 250 frames.

    :param object func: Any function (although any object will work).
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}
        self.window = None
        self.last_clear = 0

    def __call__(self, *args):
        """
        Attempts to return the results of this function call from the cache;
        if it can't find it, then it will execute the function and add it to the
        cache.
        """
        from designer import GLOBAL_DIRECTOR
        frame = GLOBAL_DIRECTOR.tick
        if GLOBAL_DIRECTOR.current_window is not self.window:
            self.window = GLOBAL_DIRECTOR.current_window
            self.cache = {}
        if frame - self.last_clear > 100:
            for key, value in list(self.cache.items()):
                data, oldframe = value
                if frame - oldframe > 250:
                    self.cache.pop(key)
            self.last_clear = frame
        try:
            data, oldframe = self.cache[args]
            self.cache[args] = (data, frame)
            return data
        except KeyError:
            res = self.func(*args)
            self.cache[args] = (res, frame)
            return res
        except TypeError as e:
            print("WARNING: Unhashable type passed to SmartMemoize."
                  "Reconsider using this decorator", e, hash(args))
            return self.func(*args)


class _ImageMemoize(SmartMemoize):
    """
    A subclass of SmartMemoize that is built explicitly for image related calls.
    It allows images to be cleared from its cache when they are updated.
    """

    def clear(self, clear_image):
        """
        Removes the given image from the cache.
        :param clear_image: The image to remove.
        :type clear_image: :class:`InternalImage <designer.core.internal_image.InternalImage>`
        """
        self.cache = dict(((image, scale) for (image, scale)
                           in self.cache.items()
                           if image is clear_image))



"""
This magic was taken from
https://code.activestate.com/recipes/81253-weakmethod/#c4

This module provides classes and methods for weakly referencing functions and
bound methods; it turns out this is a non-trivial problem.
"""

import weakref


class DeadFunctionError(Exception):
    pass


class WeakMethodBound:
    """
    Holds a weak reference to a bound method for an object.

    .. attribute::method

        The function being called.
    """

    def __init__(self, func):
        self.func = func.__func__
        self.weak_object_ref = weakref.ref(func.__self__)

    def _func(self):
        return self.func

    method = property(_func)

    def __call__(self, *arg):
        if self.weak_object_ref() is None:
            raise DeadFunctionError('Method called on dead object')
        return self.func(self.weak_object_ref(), *arg)


class WeakMethodFree:
    """
    Holds a weak reference to an unbound function. Included only for
    completeness.

    .. attribute::method

        The function being called.
    """

    def __init__(self, func):
        self.func = weakref.ref(func)

    def _func(self):
        return self.func()

    method = property(_func)

    def __call__(self, *arg):
        if self.func() is None:
            raise DeadFunctionError('Function no longer exist')
        return self.func()(arg)


def WeakMethod(func):
    """
    Attempts to create a weak reference to this function; only bound methods
    require a weakreference.
    """
    try:
        func.__func__
    except AttributeError:
        return func
    return WeakMethodBound(func)



import pygame
import math

from designer.colors import _process_color
from designer.helpers import get_width, get_height
from designer.objects.designer_object import DesignerObject
from designer.core.internal_image import InternalImage, DesignerSurface
from designer.utilities.vector import Vec2D
from designer.utilities.util import _anchor_offset


class Rectangle(DesignerObject):
    FIELDS = (*DesignerObject.FIELDS, 'color', 'border')

    def __init__(self, center, anchor, width, height, color, border):
        """
        Creates Rectangle Designer Object on window.

        :param center: x, y coordinates of center of circle
        :type center: Tuple[int]
        :param anchor: the anchor to draw the circle at
        :type anchor: str
        :param width: width of rectangle in pixels
        :type width: int
        :param height: height of rectangle in pixels
        :type height: int
        :param color: color of rectangle
        :type color: str or List[str]
        :param border: the width of the circle's line (0 is used for a filled circle)
        :type border: int
        """
        super().__init__()

        x, y = center
        x = x if x is not None else get_width() / 2
        y = y if y is not None else get_height() / 2
        center = x, y

        self._pos = center
        self._anchor = anchor
        # Rectangle specific data
        self._size = Vec2D(width, height)
        self._color = color
        self._border = border

        # And draw!
        self._redraw_internal_image()

    def __repr__(self):
        return f"<rectangle({self._color!r}, {self._size[0]}, {self._size[1]})>"

    def _recalculate_offset(self):
        size = self._size * self._scale
        offset = _anchor_offset(self._anchor, size[0], size[1])
        self._offset = Vec2D(offset) - self._transform_offset

    def _redraw_internal_image(self):
        # No flipping
        # Scaling
        width = self._size[0] * self._scale[0]
        height = self._size[1] * self._scale[1]
        size = (int(width), int(height))
        color = _process_color(self._color)
        if size[0] <= 0 or size[1] <= 0:
            target = InternalImage(size=(1, 1)).fill(color)
            self._transform_image = target._surf
            self._recalculate_offset()
            self._expire_static()
            return
        new_image = InternalImage(size=size)
        new_image.draw_rect(color, (0, 0), size, self._border or 0)
        # Rotation
        if self._angle != 0:
            old = Vec2D(new_image.rect.center)
            new_image.rotate(self._angle % 360)
            new = new_image.rect.center
            self._transform_offset = old - new
        self._transform_image = new_image._surf
        self._recalculate_offset()
        self._expire_static()

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value
        self._redraw_internal_image()

    @property
    def border(self):
        return self._border

    @border.setter
    def border(self, value):
        self._border = value
        self._redraw_internal_image()


def rectangle(color, width, height=None, x=None, y=None, anchor='center', border=None):
    '''
    Function to create a rectangle.

    :param color: color of rectangle
    :type color: str or List[str]
    :param args: left top corner of image and width and height of rectangle
    :type args: two Tuples (left, top), (width, height) or four ints left, top, width, height
    :return: Rectangle object created
    '''
    if x is not None and y is None:
        x, y = x
    elif x is None and y is None and height is not None:
        if not isinstance(height, (int, float)):
            x, y = height
            width, height = width
    elif height is None:
        width, height = width
    return Rectangle((x, y), anchor, width, height, color, border)


import pygame
import copy
import imghdr
import math

from designer.utilities.vector import Vec2D
from designer.utilities.rect import Rect
from designer.utilities.util import scale_surface


class DesignerSurface(pygame.Surface):
    """
    Internal method for creating a DesignerSurface, which is just a Pygame
    Surface with the proper full SRCALPHA, 32 bits, and alpha transparency.
    """
    def __init__(self, size):
        super().__init__((int(size[0]), int(size[1])), pygame.SRCALPHA, 32)
        self.convert_alpha()


def from_sequence(images, orientation="right", padding=0):
    """
    A function that returns a new Image from a list of images by
    placing them next to each other.

    :param images: A list of images to lay out.
    :type images: List of :class:`Image <spyral.Image>`
    :param str orientation: Either 'left', 'right', 'above', 'below', or
                            'square' (square images will be placed in a grid
                            shape, like a chess board).
    :param padding: The padding between each image. Can be specified as a
                    scalar number (for constant padding between all images)
                    or a list (for different paddings between each image).
    :type padding: int or a list of ints.
    :returns: A new :class:`Image <spyral.Image>`
    """
    if orientation == 'square':
        length = int(math.ceil(math.sqrt(len(images))))
        max_height = 0
        sequence = []
        x, y = 0, 0
        for index, image in enumerate(images):
            if index % length == 0:
                x = 0
                y += max_height
                max_height = 0
            else:
                x += image.width
                max_height = max(max_height, image.height)
            sequence.append((image, (x, y)))
    else:
        if orientation in ('left', 'right'):
            selector = Vec2D(1, 0)
        else:
            selector = Vec2D(0, 1)

        if orientation in ('left', 'above'):
            reversed(images)
            # TODO: Does this actually do anything?

        if type(padding) in (float, int):
            padding = [padding] * len(images)
        else:
            padding = list(padding)
            padding.append(0)
        base = Vec2D(0, 0)
        sequence = []
        for image, padding in zip(images, padding):
            sequence.append((image, base))
            base = base + selector * (image.size + (padding, padding))
    return from_conglomerate(sequence)


def from_conglomerate(sequence):
    """
    A function that generates a new InternalImage from a sequence of
    (image, position) pairs. These images will be placed onto a singe image
    large enough to hold all of them. More explicit and less convenient than
    :func:`from_seqeuence <spyral.image.from_sequence>`.

    :param sequence: A list of (image, position) pairs, where the positions
                     are :class:`Vec2D <spyral.Vec2D>` s.
    :type sequence: List of image, position pairs.
    :returns: A new :class:`Image <spyral.Image>`
    """
    width, height = 0, 0
    for image, (x, y) in sequence:
        width = max(width, x+image.width)
        height = max(height, y+image.height)
    new = InternalImage(size=(width, height))
    for image, (x, y) in sequence:
        new.draw_internal_image(image, (x, y))
    return new

class InternalImage(object):
    """
    The InternalImage is the basic drawable item in designer. They can be created
    either by loading from common file formats, or by creating a new
    InternalImage and using some of the draw methods. InternalImages are not drawn on
    their own, they are placed as the *internal_image* attribute on DesignerObjects to
    be drawn.

    Almost all of the methods of an InternalImage instance return the InternalImage itself,
    enabling commands to be chained in a
    `fluent interface <http://en.wikipedia.org/wiki/Fluent_interface>`_.

    :param size: If size is passed, creates a new blank internal_image of that size to
                 draw on. If you do not specify a size, you *must* pass in a
                 filename.
    :type size: :class:`Vec2D <spyral.Vec2D>`
    :param str filename:  If filename is set, the file with that name is loaded.
                          The appendix has a list of the
                          :ref:`valid image formats<ref.image_formats>`. If you do
                          not specify a filename, you *must* pass in a size.

    """

    def __init__(self, filename=None, size=None, fileobj=None):
        if size is not None and filename is not None:
            raise ValueError("Must specify exactly one of size and filename. See http://platipy.org/en/latest/spyral_docs.html#spyral.internal_image.InternalImage")
        if size is None and filename is None:
            raise ValueError("Must specify exactly one of size and filename. See http://platipy.org/en/latest/spyral_docs.html#spyral.internal_image.InternalImage")

        if size is not None:
            self._surf = DesignerSurface(size)
            self._name = None
        else:
            self._name = filename
            if fileobj is None:
                fileobj = filename
            self._surf = pygame.image.load(fileobj).convert_alpha()
        self._version = 1

    def _get_width(self):
        return self._surf.get_width()

    #: The width of this internal_image in pixels (int). Read-only.
    width = property(_get_width)

    def _get_height(self):
        return self._surf.get_height()

    #: The height of this internal_image in pixels (int). Read-only.
    height = property(_get_height)

    def _get_size(self):
        return Vec2D(self._surf.get_size())

    #: The (width, height) of the internal_image (:class:`Vec2D <spyral.Vec2D`).
    #: Read-only.
    size = property(_get_size)

    @property
    def rect(self):
        return self._surf.get_rect()

    def fill(self, color):
        """
        Fills the entire internal_image with the specified color.

        :param color: a three-tuple of RGB values ranging from 0-255. Example:
                      (255, 128, 0) is orange.
        :type color: a three-tuple of ints.
        :returns: This internal_image.
        """
        self._surf.fill(color)
        self._version += 1
        scale_surface.clear(self._surf)
        return self

    def draw_rect(self, color, position, size=None,
                  border_width=0, anchor='topleft'):
        """
        Draws a rectangle on this internal_image.

        :param color: a three-tuple of RGB values ranging from 0-255. Example:
                      (255, 128, 0) is orange.
        :type color: a three-tuple of ints.
        :param position: The starting position of the rect (top-left corner). If
                         position is a Rect, then size should be `None`.
        :type position: :class:`Vec2D <spyral.Vec2D>` or
                        :class:`Rect <spyral.Rect>`
        :param size: The size of the rectangle; should not be given if position
                     is a rect.
        :type size: :class:`Vec2D <spyral.Vec2D>`
        :param int border_width: The width of the border to draw. If it is 0,
                                 the rectangle is filled with the color
                                 specified.
        :param str anchor: The anchor parameter is an
                           :ref:`anchor position <ref.anchors>`.
        :returns: This internal_image.
        """
        if size is None:
            rect = Rect(position)
        else:
            rect = Rect(position, size)
        offset = self._calculate_offset(anchor, rect.size)
        pygame.draw.rect(self._surf, color,
                             (rect.pos + offset, rect.size), border_width)
        self._version += 1
        scale_surface.clear(self._surf)
        return self

    def draw_lines(self, color, points, width=1, closed=False):
        """
        Draws a series of connected lines on a internal_image, with the
        vertices specified by points. This does not draw any sort of
        end caps on lines.

        :param color: a three-tuple of RGB values ranging from 0-255. Example:
                      (255, 128, 0) is orange.
        :type color: a three-tuple of ints.
        :param points: A list of points that will be connected, one to another.
        :type points: A list of :class:`Vec2D <spyral.Vec2D>` s.
        :param int width: The width of the lines.
        :param bool closed: If closed is True, the first and last point will be
                            connected. If closed is True and width is 0, the
                            shape will be filled.
        :returns: This internal_image.
        """
        if width == 1:
            pygame.draw.aalines(self._surf, color, closed, points)
        else:
            pygame.draw.lines(self._surf, color, closed, points, width)
        self._version += 1
        scale_surface.clear(self._surf)
        return self

    def draw_circle(self, color, position, radius, width=0, anchor='topleft'):
        """
        Draws a circle on this internal_image.

        :param color: a three-tuple of RGB values ranging from 0-255. Example:
                      (255, 128, 0) is orange.
        :type color: a three-tuple of ints.
        :param position: The center of this circle
        :type position: :class:`Vec2D <spyral.Vec2D>`
        :param int radius: The radius of this circle
        :param int width: The width of the circle. If it is 0, the circle is
                          filled with the color specified.
        :param str anchor: The anchor parameter is an
                           :ref:`anchor position <ref.anchors>`.
        :returns: This internal_image.
        """
        offset = self._calculate_offset(anchor)
        pygame.draw.circle(self._surf, color, (position + offset).floor(),
                           radius, width)
        self._version += 1
        scale_surface.clear(self._surf)
        return self

    def draw_ellipse(self, color, position, size=None,
                     border_width=0, anchor='topleft'):
        """
        Draws an ellipse on this internal_image.

        :param color: a three-tuple of RGB values ranging from 0-255. Example:
                      (255, 128, 0) is orange.
        :type color: a three-tuple of ints.
        :param position: The starting position of the ellipse (top-left corner).
                         If position is a Rect, then size should be `None`.
        :type position: :class:`Vec2D <spyral.Vec2D>` or
                        :class:`Rect <spyral.Rect>`
        :param size: The size of the ellipse; should not be given if position is
                     a rect.
        :type size: :class:`Vec2D <spyral.Vec2D>`
        :param int border_width: The width of the ellipse. If it is 0, the
                          ellipse is filled with the color specified.
        :param str anchor: The anchor parameter is an
                           :ref:`anchor position <ref.anchors>`.
        :returns: This internal_image.
        """
        if size is None:
            rect = Rect(position)
        else:
            rect = Rect(position, size)
        offset = self._calculate_offset(anchor, rect.size)
        pygame.draw.ellipse(self._surf, color,
                            (rect.pos + offset, rect.size), border_width)
        self._version += 1
        scale_surface.clear(self._surf)
        return self

    def draw_point(self, color, position, anchor='topleft'):
        """
        Draws a point on this internal_image.

        :param color: a three-tuple of RGB values ranging from 0-255. Example:
                      (255, 128, 0) is orange.
        :type color: a three-tuple of ints.
        :param position: The position of this point.
        :type position: :class:`Vec2D <spyral.Vec2D>`
        :param str anchor: The anchor parameter is an
                           :ref:`anchor position <ref.anchors>`.
        :returns: This internal_image.
        """
        offset = self._calculate_offset(anchor)
        self._surf.set_at(position + offset, color)
        self._version += 1
        scale_surface.clear(self._surf)
        return self

    def draw_arc(self, color, start_angle, end_angle,
                 position, size=None, border_width=0, anchor='topleft'):
        """
        Draws an elliptical arc on this internal_image.

        :param color: a three-tuple of RGB values ranging from 0-255. Example:
                      (255, 128, 0) is orange.
        :type color: a three-tuple of ints.
        :param float start_angle: The starting angle, in radians, of the arc.
        :param float end_angle: The ending angle, in radians, of the arc.
        :param position: The starting position of the ellipse (top-left corner).
                         If position is a Rect, then size should be `None`.
        :type position: :class:`Vec2D <spyral.Vec2D>` or
                        :class:`Rect <spyral.Rect>`
        :param size: The size of the ellipse; should not be given if position is
                     a rect.
        :type size: :class:`Vec2D <spyral.Vec2D>`
        :param int border_width: The width of the ellipse. If it is 0, the
                          ellipse is filled with the color specified.
        :param str anchor: The anchor parameter is an
                           :ref:`anchor position <ref.anchors>`.
        :returns: This internal_image.
        """
        if size is None:
            rect = Rect(position)
        else:
            rect = Rect(position, size)
        offset = self._calculate_offset(anchor, rect.size)
        pygame.draw.arc(self._surf, color, (rect.pos + offset, rect.size),
                        start_angle, end_angle, border_width)
        self._version += 1
        scale_surface.clear(self._surf)
        return self

    def draw_internal_image(self, internal_image, position=(0, 0), anchor='topleft'):
        """
        Draws another internal_image over this one.

        :param internal_image: The internal_image to overlay on top of this one.
        :type internal_image: :class:`InternalImage <designer.core.internal_image.InternalImage>`
        :param position: The position of this internal_image.
        :type position: :class:`Vec2D <spyral.Vec2D>`
        :param str anchor: The anchor parameter is an
                           :ref:`anchor position <ref.anchors>`.
        :returns: This internal_image.
        """
        offset = self._calculate_offset(anchor, internal_image._surf.get_size())
        self._surf.blit(internal_image._surf, position + offset)
        self._version += 1
        scale_surface.clear(self._surf)
        return self

    def draw_surface(self, surf, position=(0, 0), anchor='topleft'):
        """
        Draws a pygame surface over this one.

        :param internal_image: The internal_image to overlay on top of this one.
        :type internal_image: :class:`InternalImage <designer.core.internal_image.InternalImage>`
        :param position: The position of this internal_image.
        :type position: :class:`Vec2D <spyral.Vec2D>`
        :param str anchor: The anchor parameter is an
                           :ref:`anchor position <ref.anchors>`.
        :returns: This internal_image.
        """
        offset = self._calculate_offset(anchor, surf.get_size())
        self._surf.blit(surf, position + offset)
        self._version += 1
        scale_surface.clear(self._surf)
        return self

    def rotate(self, angle):
        """
        Rotates the internal_image by angle degrees clockwise. This may change the internal_image
        dimensions if the angle is not a multiple of 90.

        Successive rotations degrate internal_image quality. Save a copy of the
        original if you plan to do many rotations.

        :param float angle: The number of degrees to rotate.
        :returns: This internal_image.
        """
        self._surf = pygame.transform.rotate(self._surf, angle).convert_alpha()
        self._version += 1
        return self

    def scale(self, size):
        """
        Scales the internal_image to the destination size.

        :param size: The new size of the internal_image.
        :type size: :class:`Vec2D <spyral.Vec2D>`
        :returns: This internal_image.
        """
        self._surf = pygame.transform.smoothscale(self._surf,
                                                  size).convert_alpha()
        self._version += 1
        return self

    def flip(self, flip_x=True, flip_y=True):
        """
        Flips the internal_image horizontally, vertically, or both.

        :param bool flip_x: whether to flip horizontally.
        :param bool flip_y: whether to flip vertically.
        :returns: This internal_image.
        """
        self._version += 1
        self._surf = pygame.transform.flip(self._surf,
                                           flip_x, flip_y).convert_alpha()
        return self

    @classmethod
    def from_surface(cls, surf: pygame.Surface):
        image = InternalImage(size=surf.get_size())
        return image.draw_surface(surf)

    def copy(self):
        """
        Returns a copy of this internal_image that can be changed while preserving the
        original.

        :returns: A new internal_image.
        """
        new = copy.copy(self)
        new._surf = self._surf.copy()
        return new

    def crop(self, position, size=None):
        """
        Removes the edges of an internal_image, keeping the internal rectangle specified
        by position and size.

        :param position: The upperleft corner of the internal rectangle that
                         will be preserved.
        :type position: a :class:`Vec2D <spyral.Vec2D>` or a
                        :class:`Rect <spyral.Rect>`.
        :param size: The size of the internal rectangle to preserve. If a Rect
                     was passed in for position, this should be None.
        :type size: :class:`Vec2D <spyral.Vec2D>` or None.
        :returns: This internal_image.
        """
        if size is None:
            rect = Rect(position)
        else:
            rect = Rect(position, size)
        new = DesignerSurface(size)
        new.blit(self._surf, (0, 0), (rect.pos, rect.size))
        self._surf = new
        self._version += 1
        return self

    def _calculate_offset(self, anchor_type, size=(0, 0)):
        """
        Internal method for calculating the offset associated with an
        anchor type.

        :param anchor_type: A string indicating the position of the anchor,
                            taken from :ref:`anchor position <ref.anchors>`. A
                            numerical offset can also be specified.
        :type anchor_type: str or a :class:`Vec2D <spyral.Vec2D>`.
        :param size: The size of the region to offset in.
        :type size: :class:`Vec2D <spyral.Vec2D>`.
        """
        w, h = self._surf.get_size()
        w2, h2 = size

        if anchor_type == 'topleft':
            return Vec2D(0, 0)
        elif anchor_type == 'topright':
            return Vec2D(w - w2, 0)
        elif anchor_type == 'midtop':
            return Vec2D((w - w2) / 2., 0)
        elif anchor_type == 'bottomleft':
            return Vec2D(0, h - h2)
        elif anchor_type == 'bottomright':
            return Vec2D(w - w2, h - h2)
        elif anchor_type == 'midbottom':
            return Vec2D((w - w2) / 2., h - h2)
        elif anchor_type == 'midleft':
            return Vec2D(0, (h - h2) / 2.)
        elif anchor_type == 'midright':
            return Vec2D(w - w2, (h - h2) / 2.)
        elif anchor_type == 'center':
            return Vec2D((w - w2) / 2., (h - h2) / 2.)
        else:
            return Vec2D(anchor_type) - Vec2D(w2, h2)



import imghdr
from typing import Optional, Union

from urllib.request import urlopen, Request
import pygame
import math
import sys
import os
import io

import designer
from designer.core.event import register, KNOWN_EVENTS
from designer.core.director import Director
from designer.core.internal_image import InternalImage
from designer.utilities.vector import Vec2D
from designer.utilities.argument_checks import make_suggestions


def check_initialized(**kwargs):
    """
    Checks if global state exists and creates one if it does not.

    :return: bool Indicates whether a new director got set up
    """

    if not designer.GLOBAL_DIRECTOR:
        designer.GLOBAL_DIRECTOR = Director(**kwargs)
        designer.GLOBAL_DIRECTOR._setup_initial_window()
        return True
    return False


def draw(*objs):
    """
    Draws Designer Objects on window.

    :param objs: objects that have been created to draw on the window
    :type objs: DesignerObjects

    :return: None
    """
    check_initialized()
    designer.GLOBAL_DIRECTOR.start(objs)


def start(*objs):
    check_initialized()
    if len(objs) == 1:
        objs = objs[0]
    else:
        objs = list(objs)
    designer.GLOBAL_DIRECTOR.start(objs)


def stop():
    """ Stops the game. """
    check_initialized()
    designer.GLOBAL_DIRECTOR.stop()


def pause():
    """ Pauses event processing in the game. """
    check_initialized()
    designer.GLOBAL_DIRECTOR.pause()


def set_window_color(color):
    """
    Changes window color to given color.
    Must call before adding any DesignerObjects.

    :param color: color to change window to
    :type color: str or List[str]

    :return: None
    """
    check_initialized()
    designer.GLOBAL_DIRECTOR.window_color = color


def get_window_color():
    check_initialized()
    return designer.GLOBAL_DIRECTOR.window_color


def set_window_title(caption: str):
    """
    Set the title of the game's window (usually the title of your game).

    :param caption: The caption that will be displayed in the window.
                    Typically the name of your game.
    :type caption: ``str``
    """
    check_initialized()
    designer.GLOBAL_DIRECTOR.window_title = caption


def get_window_title() -> str:
    check_initialized()
    return designer.GLOBAL_DIRECTOR.window_title


def set_game_state(new_state):
    check_initialized()
    designer.GLOBAL_DIRECTOR._game_state = new_state


def set_world_state(new_state):
    set_game_state(new_state)


def set_window_state(new_state):
    set_game_state(new_state)


def set_window_size(width, height):
    """
    Set size of window in pixels.
    Must call before adding any DesignerObjects.

    :param width: number of pixels to set horizontal size of window
    :type width: int
    :param height: number of pixels to set vertical size of window
    :type height: int
    :return: None
    """
    if not check_initialized(width=width, height=height):
        designer.GLOBAL_DIRECTOR.screen = pygame.display.set_mode(size=(width, height))
    designer.GLOBAL_DIRECTOR.window_size = width, height


def get_width():
    """
    Get the width of the window.

    :return: pixels of horizontal width of window
    :rtype: int
    """
    check_initialized()
    return designer.GLOBAL_DIRECTOR.window_size[0]


def get_height():
    """
    Get the height of the window.

    :return: pixels of vertical height of window
    :rtype: int
    """

    check_initialized()
    return designer.GLOBAL_DIRECTOR.window_size[1]


def get_mouse_cursor():
    check_initialized()
    return designer.GLOBAL_DIRECTOR.mouse.cursor


def set_mouse_cursor(value):
    check_initialized()
    designer.GLOBAL_DIRECTOR.mouse.cursor = value


def get_mouse_visible():
    check_initialized()
    return designer.GLOBAL_DIRECTOR.mouse.visible


def set_mouse_visible(value):
    check_initialized()
    designer.GLOBAL_DIRECTOR.mouse.visible = value


def get_keyboard_delay():
    check_initialized()
    return designer.GLOBAL_DIRECTOR.keyboard.delay


def set_keyboard_delay(value):
    check_initialized()
    designer.GLOBAL_DIRECTOR.keyboard.delay = value


def get_keyboard_interval():
    check_initialized()
    return designer.GLOBAL_DIRECTOR.keyboard.interval


def set_keyboard_interval(value):
    check_initialized()
    designer.GLOBAL_DIRECTOR.keyboard.interval = value


def get_keyboard_repeat():
    check_initialized()
    return designer.GLOBAL_DIRECTOR.keyboard.repeat


def set_keyboard_repeat(value):
    check_initialized()
    designer.GLOBAL_DIRECTOR.keyboard.repeat = value


def enable_keyboard_repeating():
    check_initialized()
    designer.GLOBAL_DIRECTOR.keyboard.repeat = True
    designer.GLOBAL_DIRECTOR.keyboard.delay = 1


def disable_keyboard_repeating():
    check_initialized()
    designer.GLOBAL_DIRECTOR.keyboard.repeat = designer.GLOBAL_DIRECTOR.keyboard.DEFAULT_REPEAT
    designer.GLOBAL_DIRECTOR.keyboard.delay = designer.GLOBAL_DIRECTOR.keyboard.DEFAULT_DELAY


def background_music(filename, volume=1.0, loop=True):
    check_initialized()
    designer.GLOBAL_DIRECTOR.music.filename = filename
    designer.GLOBAL_DIRECTOR.music.play(loop=-1 if loop is True else loop)
    designer.GLOBAL_DIRECTOR.music.volume = volume


def play_music(filename, volume=1.0, loop=True):
    return background_music(filename, volume, loop)


def pause_music():
    check_initialized()
    designer.GLOBAL_DIRECTOR.music.pause()


def set_music_volume(volume):
    check_initialized()
    designer.GLOBAL_DIRECTOR.music.volume = volume


def is_music_playing():
    check_initialized()
    return designer.GLOBAL_DIRECTOR.music.playing


def get_music_volume():
    check_initialized()
    return designer.GLOBAL_DIRECTOR.music.volume


def play_sound(path, volume=1.0):
    check_initialized()
    designer.GLOBAL_DIRECTOR.sfx.play(path, volume)


def stop_music():
    check_initialized()
    designer.GLOBAL_DIRECTOR.music.stop()


def rewind_music():
    check_initialized()
    designer.GLOBAL_DIRECTOR.music.rewind()


def continue_music():
    check_initialized()
    designer.GLOBAL_DIRECTOR.music.unpause()


def set_music_position(time):
    check_initialized()
    designer.GLOBAL_DIRECTOR.music.set_time_position(time)


def get_music_position():
    check_initialized()
    return designer.GLOBAL_DIRECTOR.music.get_time_position()


def when(event: Union[str, callable], *funcs):
    check_initialized()
    if callable(event):
        event_function = event
        event = 'updating'

        def _dynamic_event(func):
            def _inner_dynamic_event(event, world):
                if event_function(world):
                    window = designer.GLOBAL_DIRECTOR.current_window
                    window._send_event_to_handler(event, 'updating', func, None, None, None, None)

            return _inner_dynamic_event

        funcs = [_dynamic_event(f) for f in funcs]
    if event not in KNOWN_EVENTS and not any(e.startswith(event) for e in KNOWN_EVENTS):
        suggestions = make_suggestions(event, KNOWN_EVENTS)
        if suggestions:
            raise ValueError(f"Unrecognized event {event!r}. Perhaps you meant one of: {suggestions}")
        else:
            raise ValueError(
                f"Unrecognized event {event!r}. Check the documentation to see possible events (like 'updating' and 'starting').")
    for func in funcs:
        register(event, func)


def colliding(*args):
    # TODO: Could use collide_circle and collide_mask for improved collisions
    check_initialized()
    if len(args) == 2:
        obj1 = args[0]
        obj2 = args[1]
        return obj1.collide_other(obj2)
    elif len(args) == 3:
        if isinstance(args[0], (int, float)) and isinstance(args[1], (int, float)):
            x, y, obj = args
        elif isinstance(args[1], (int, float)) and isinstance(args[2], (int, float)):
            obj, x, y = args
        else:
            raise ValueError(f"Could not recognize the type of the arguments: {args!r}")
        return obj.collide_point(Vec2D(x, y))


def destroy(*gobjects):
    check_initialized()
    for gobject in gobjects:
        gobject.destroy()


# TODO: Fix to centralize
_USER_AGENT = "Designer Game Library for Python"


def background_image(path):
    check_initialized()
    try:
        path_strs = path.split('/')
        fixed_paths = os.path.join(*path_strs)
        if os.path.exists(fixed_paths):
            designer.GLOBAL_DIRECTOR.current_window.background = InternalImage(fixed_paths)
        else:
            raise FileNotFoundError(fixed_paths)
    except FileNotFoundError as err:
        try:
            req = Request(path, headers={'User-Agent': _USER_AGENT})
            with urlopen(req) as opened_image:
                image_str = opened_image.read()
                image_file = io.BytesIO(image_str)
                designer.GLOBAL_DIRECTOR.current_window.background = InternalImage(filename=path, fileobj=image_file)
        except:
            print(f"Unexpected error while loading background image: {path}\n", sys.exc_info()[0])
            raise


def get_director():
    check_initialized()
    return designer.GLOBAL_DIRECTOR



from os import environ

environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'

import pygame
from designer.core.director import *
from designer.core.event import *
from designer.helpers import *
from designer.animation import *
from designer.utilities.easings import *
from designer.objects import *
from designer.colors import *
from designer.positioning import *
from designer.keyboard import *
from designer.mouse import *

GLOBAL_DIRECTOR: Director = None

__all__ = [
    'circle',
    'ellipse',
    'arc',
    'line',
    'rectangle',
    'text',
    'shape',
    'background_image',
    'image',
    'group',
    'draw',
    'set_window_color',
    'set_window_size',
    'get_height',
    'get_width',
    'when',
    'start',
    'stop',
    'pause',
    'colliding',
    'destroy',
    'DesignerObject',
    'above', 'below',
    # Director stuff
    'get_director',
    # Window stuff
    'set_window_title', 'get_window_title',
    # Keyboard stuff
    'get_keyboard_repeat', 'set_keyboard_repeat',
    'get_keyboard_delay', 'set_keyboard_delay',
    'get_keyboard_interval', 'set_keyboard_interval',
    'enable_keyboard_repeating', 'disable_keyboard_repeating',
    # Mouse stuff
    'get_mouse_cursor', 'set_mouse_cursor',
    'get_mouse_visible', 'set_mouse_visible',
    'get_mouse_position', 'set_mouse_position',
    'get_mouse_x', 'get_mouse_y',
    # Animations
    'Animation', 'linear_animation', 'sequence_animation',
    'glide_around',
    'glide_right',
    'glide_left',
    'glide_up',
    'glide_down',
    'glide_in_degrees',
    'spin',
    # Easings
    'Linear', 'Iterate',
    # Music
    'play_sound',
    'play_music', 'background_music', 'pause_music', 'set_music_volume', 'is_music_playing',
    'get_music_volume', 'stop_music', 'rewind_music', 'continue_music', 'set_music_position', 'get_music_position'
]



import math
from typing import List

import pygame

import designer
from designer.core.internal_image import DesignerSurface, InternalImage
from designer.colors import _process_color
from designer.helpers import get_width, get_height
from designer.objects.designer_object import DesignerObject
from designer.utilities.vector import Vec2D
from designer.utilities.util import rect_from_points, _anchor_offset
from designer.utilities.argument_checks import is_non_empty_iterable_of_points, at_least_three


class Shape(DesignerObject):
    DEFAULT_BORDER_WIDTH = 1
    FIELDS = (*DesignerObject.FIELDS, 'points', 'color', 'border')

    def __init__(self, center, points, anchor, color, border):

        super().__init__()

        if center in (None, (None, None)):
            center = points[0]
        if center in (None, (None, None)):
            x, y = Vec2D(get_width() / 2, get_height() / 2)
            center = x, y

        self._pos = center
        self._anchor = anchor
        # Polygon specific data
        self._border = border
        self._points: List[Vec2D] = [Vec2D(p) for p in points]
        self._size = self._get_bounds() # computed
        self._color = color

        # And Draw!
        self._redraw_internal_image()

    def _get_bounds(self):
        if not self._points:
            return Vec2D(1, 1)
        bounds = rect_from_points(self._points)
        # TODO: Refine this calculation of border expansion
        if self._border > 1:
            bounds.inflate_ip(self._border*2, self._border*2)
        return bounds.size

    def _recalculate_offset(self):
        if self._transform_image is None:
            return
        size = self._scale * self._transform_image.get_size()
        offset = _anchor_offset(self._anchor, size[0], size[1])
        self._offset = Vec2D(offset) - self._transform_offset

    def _redraw_internal_image(self):
        target = InternalImage(size=self._size)
        if not self._points or len(self._points) < 3:
            return self._make_blank_surface()
        origin = self._points[0]
        # TODO: Refine this calculation of border expansion
        if self._border > 1:
            origin = origin + Vec2D(self._border, self._border)
        offset_points = [point-origin for point in self._points]
        pygame.draw.polygon(target._surf, _process_color(self._color),
                            offset_points, self._border)
        self._default_redraw_transforms(target)

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value
        self._redraw_internal_image()

    @property
    def border(self):
        return self._border

    @border.setter
    def border(self, value):
        self._border = value
        self._size = self._get_bounds()
        self._redraw_internal_image()

    @property
    def points(self):
        return self._points

    @points.setter
    def points(self, value):
        if not is_non_empty_iterable_of_points(value) or len(value) < 3:
            raise ValueError("The points of a shape must be a list of at least 3 points")
        self._points = value
        self._size = self._get_bounds()
        self._redraw_internal_image()


def shape(color, *points, x=None, y=None, anchor='topleft', border=None, filled=True):
    """
    Function to create a shape of at least three points.

    :param color: color of shape.
    :type color: str
    :param points: coordinates of points of shape
    :type points: List[Tuple] in (x, y), (x, y) format of points to be connected of shape
    :return: Shape object created
    """
    if len(points) == 1 and isinstance(points, (list, tuple)):
        points = points[0]
    if not is_non_empty_iterable_of_points(points):
        raise ValueError("The points of a shape must be a list of at least 3 points.")
    if not at_least_three(points):
        raise ValueError("You need at least three points in your shape.")
    # If not at origin, then translate everything and use x/y as start
    if x is None and y is None:
        if points[0] != Vec2D(0, 0):
            x, y = points[0]
            points = [-Vec2D(x, y)+p for p in points]
    # If x/y are given at this point, they are the start
    if x is not None and y is not None:
        center = Vec2D(x, y)
    else:
        # otherwise the first point is
        center = points[0]
    # And handle filling
    if filled is True:
        border = 0
    elif filled is False:
        border = border or Shape.DEFAULT_BORDER_WIDTH
    elif border is None:
        border = Shape.DEFAULT_BORDER_WIDTH
    return Shape(center, points, anchor, color, border)



"""The mouse modules provides an interface to adjust the mouse cursor.

.. attribute:: visible

    `Bool` that adjust whether the mouse cursor should be shown. This is useful
    if you want to, for example, use a Sprite instead of the regular mouse
    cursor.

.. attribute:: cursor

    `str` value that lets you choose from among the built-in options for
    cursors. The options are:

        * ``"arrow"`` : the regular arrow-shaped cursor
        * ``"diamond"`` : a diamond shaped cursor
        * ``"x"`` : a broken X, useful for indicating disabled states.
        * ``"left"``: a triangle pointing to the left
        * ``"right"``: a triangle pointing to the right

    .. warning:: Custom non-Sprite mouse cursors are currently not supported.

"""
import pygame

cursors = {"arrow": pygame.cursors.arrow,
           "diamond": pygame.cursors.diamond,
           "x": pygame.cursors.broken_x,
           "left": pygame.cursors.tri_left,
           "right": pygame.cursors.tri_right}


class MouseModule:
    def __init__(self):
        self._visible = True
        self._cursor = 'arrow'

    @property
    def cursor(self):
        return self._cursor

    @cursor.setter
    def cursor(self, value):
        if value in cursors:
            self._cursor = cursors[value]
        else:
            self._cursor = value
        pygame.mouse.set_cursor(*self._cursor)

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, value):
        pygame.mouse.set_visible(value)
        self._visible = value


def get_mouse_position():
    return pygame.mouse.get_pos()

def set_mouse_position(x, y):
    pygame.mouse.set_pos((x, y))

def get_mouse_x():
    return get_mouse_position()[0]

def get_mouse_y():
    return get_mouse_position()[1]



import imghdr
from typing import Optional

from urllib.request import urlopen, Request
import pygame
import math
import sys
import os
import io

from designer.colors import _process_color
from designer.helpers import get_width, get_height
from designer.objects.designer_object import DesignerObject
from designer.core.internal_image import InternalImage, DesignerSurface
from designer.utilities import Vec2D
from designer.utilities.util import _anchor_offset
from designer.utilities.gif_image import GifImage


class Image(DesignerObject):
    _USER_AGENT = "Designer Game Library for Python"
    FIELDS = (*DesignerObject.FIELDS, 'filename', 'image')
    _IMAGE_CACHE = {}
    _GIF_CACHE = {}

    def __init__(self, center, path, anchor, **kwargs):
        """
        Creates Image Designer Object on window

        :param center: x, y coordinates of center of circle
        :type center: Tuple[int]
        :param anchor: the anchor to draw the circle at
        :type anchor: str
        :param path: either url or local file path to image to load on screen
        :type path: str
        :param width: width of image in pixels
        :type width: int
        :param height: height of image in pixels
        :type height: int
        """
        super().__init__()

        x, y = center
        x = x if x is not None else get_width() / 2
        y = y if y is not None else get_height() / 2
        center = x, y

        self._pos = center
        self._anchor = anchor
        # Image specific data
        self._filename = path
        #: Internal field holding the original version of the image
        self._internal_image: Optional[InternalImage] = None
        self._internal_image_version: Optional[int] = None

        self._load_image()

        for key, value in kwargs.items():
            self[key] = value

        # And draw!
        self._redraw_internal_image()
        self._recalculate_offset()

    def _load_image(self):
        if self._filename in self._GIF_CACHE:
            # TODO: Return a copy instead?
            return self.animate(self._GIF_CACHE[self._filename])
        if self._filename in self._IMAGE_CACHE:
            self._internal_image = self._IMAGE_CACHE[self._filename].copy()
        try:
            path_strs = self._filename.split('/')
            fixed_paths = os.path.join(*path_strs)
            if os.path.exists(fixed_paths):
                if imghdr.what(fixed_paths) == 'gif':
                    with open(fixed_paths) as image_file:
                        gif = GifImage(image_file)
                        self.animate(gif)
                        self._GIF_CACHE[fixed_paths] = gif
                else:
                    self._internal_image = InternalImage(fixed_paths)
                    self._IMAGE_CACHE[fixed_paths] = self._internal_image
            else:
                raise FileNotFoundError(fixed_paths)
        except FileNotFoundError as err:
            try:
                req = Request(self._filename, headers={'User-Agent': self._USER_AGENT})
                with urlopen(req) as opened_image:
                    image_str = opened_image.read()
                    image_file = io.BytesIO(image_str)
                    if imghdr.what(None, h=image_str) == 'gif':
                        gif = GifImage(image_file)
                        self.animate(gif)
                        self._GIF_CACHE[self._filename] = gif
                    else:
                        self._internal_image = InternalImage(filename=self._filename, fileobj=image_file)
                        self._IMAGE_CACHE[self._filename] = self._internal_image
            except:
                if self._filename.startswith('https://') or self._filename.startswith('http://'):
                    raise ValueError(f"Unexpected error while loading url: {self._filename!r}")
                raise ValueError(f"Unexpected error while loading image from filename: {self._filename!r}")

    def _recalculate_offset(self):
        """
        Recalculates this designer object's offset based on its position, transform
        offset, anchor, its image, and the image's scaling.
        """
        if self._internal_image is None:
            return
        size = self._scale * self._internal_image.size
        offset = _anchor_offset(self._anchor, size[0], size[1])
        self._size = size
        self._offset = Vec2D(offset) - self._transform_offset

    def _redraw_internal_image(self):
        """
        Calculates the transforms that need to be applied to this designer object's
        image. In order: flipping, scaling, and rotation.
        """
        if self._internal_image is None:
            return
        source = self._internal_image._surf
        # Flip
        if self._flip_x or self._flip_y:
            source = pygame.transform.flip(source, self._flip_x, self._flip_y)
        # Scale
        if self._scale != (1.0, 1.0):
            new_size = self._scale * self._internal_image.size
            new_size = (int(new_size[0]), int(new_size[1]))
            if 0 in new_size:
                return self._make_blank_surface()
            new_surf = DesignerSurface(new_size)
            source = pygame.transform.smoothscale(source, new_size, new_surf)
        # Rotate
        if self._angle != 0:
            angle = self._angle % 360
            old = Vec2D(source.get_rect().center)
            source = pygame.transform.rotate(source, angle).convert_alpha()
            new = source.get_rect().center
            self._transform_offset = old - new
        # Finish updates
        self._transform_image = source
        self._recalculate_offset()
        self._expire_static()

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        if value == self._filename:
            return
        self._filename = value
        self._load_image()
        self._redraw_internal_image()
        self._recalculate_offset()
        self._expire_static()

    @property
    def image(self):
        return self._internal_image

    @image.setter
    def image(self, value):
        if isinstance(value, DesignerObject):
            self._internal_image = value._transform_image.copy()
            self._filename = None
            self._redraw_internal_image()
            self._recalculate_offset()
            self._expire_static()
            return
        if isinstance(value, pygame.Surface):
            value = InternalImage.from_surface(value)
        if not isinstance(value, InternalImage):
            return
        if value == self._internal_image and self._internal_image._version == value._version:
            return
        self._internal_image = value
        self._filename = value._name
        self._redraw_internal_image()
        self._recalculate_offset()
        self._expire_static()


def image(path, x=None, y=None, anchor='center', **kwargs):
    """
    Function to create an image.

    :param path: local file path or url of image to upload
    :type path: str
    :param args: left top corner of image and width and height of image
    :type args: two Tuples (left, top), (width, height) or four ints left, top, width, height
    :return: Image object
    """
    if x is not None and y is None:
        x, y = x
    return Image((x, y), path, anchor, **kwargs)




"""
This module provides a set of built-in easings which can be used by any
game. Additionally, custom easings can be built. An easing should be a
function (or callable) which takes in a sprite, and a time delta which
is normalized to [0,1], and returns the state of easing at that time.
See the source code of this module for some example implementations.
Built-in easings are stateless, so the same animation can be used many
times or on many different objects. Custom easings do not have to be
stateless.

Visualizations of these easings are available at
`http://easings.net <http://easings.net>`_ .
"""

import math
import random


def Linear(start=0.0, finish=1.0):
    """
    Linearly increasing: f(x) = x
    """

    def linear_easing(sprite, delta):
        return (finish - start) * (delta) + start

    return linear_easing


def QuadraticIn(start=0.0, finish=1.0):
    """
    Quadratically increasing, starts slower : f(x) = x ^ 2
    """

    def quadratic_easing(sprite, delta):
        return start + (finish - start) * delta * delta

    return quadratic_easing


def QuadraticOut(start=0.0, finish=1.0):
    """
    Quadratically increasing, starts faster : f(x) = 2x - x^2
    """

    def quadratic_out_easing(sprite, delta):
        return start + (finish - start) * (2.0 * delta - delta * delta)

    return quadratic_out_easing


def QuadraticInOut(start=0.0, finish=1.0):
    """
    Quadratically increasing, starts and ends slowly but fast in the middle.
    """

    def quadratic_in_out_easing(sprite, delta):
        delta *= 2
        if delta < 1:
            return start + 0.5 * delta * delta * (finish - start)
        delta -= 1
        return start + (delta - 0.5 * delta * delta + 0.5) * (finish - start)

    return quadratic_in_out_easing


def CubicIn(start=0.0, finish=1.0):
    """
    Cubically increasing, starts very slow : f(x) = x^3
    """

    def cubic_in_easing(sprite, delta):
        return start + (delta * delta * delta) * (finish - start)

    return cubic_in_easing


def CubicOut(start=0.0, finish=1.0):
    """
    Cubically increasing, starts very fast : f(x) = 1 + (x-1)^3
    """

    def cubic_out_easing(sprite, delta):
        delta -= 1.0
        return start + (delta * delta * delta + 1.0) * (finish - start)

    return cubic_out_easing


def CubicInOut(start=0.1, finish=1.0):
    """
    Cubically increasing, starts and ends very slowly but very fast in the
    middle.
    """

    def cubic_in_out_easing(sprite, delta):
        delta *= 2.0
        if delta < 1.0:
            return start + 0.5 * delta * delta * delta * (finish - start)
        delta -= 2.0
        return ((1.0 + 0.5 * delta * delta * delta) *
                (finish - start) +
                2.0 * start)

    return cubic_in_out_easing


def Iterate(items, times=1):
    """
    Iterate over a list of items. This particular easing is very useful
    for creating image animations, e.g.::

        walk_images = [spyral.Image('f1.png'), spyral.Image('f2.png'), spyral.Image('f3.png')]
        walking_animation = Animation('image', easing.Iterate(walk_images), 2.0, loop=True)
        my_sprite.animate(walking_animation)

    :param list items: A list of items (e.g., a list of
                       :class:`Images <spyral.Image>`).
    :param int times: The number of times to iterate through the list.
    """

    def iterate_easing(sprite, delta):
        # We preturb the result slightly negative so that it ends on
        # the last frame instead of looping back to the first
        i = round(delta * len(items) * times)
        return items[int(i % len(items))]

    return iterate_easing


def Sine(amplitude=1.0, phase=0, end_phase=2.0 * math.pi):
    """
    Depending on the arguments, moves at a different pace according to the sine
    function.
    """

    def sin_easing(sprite, delta):
        return amplitude * math.sin(phase + delta * (2.0 * math.pi - phase))

    return sin_easing


def LinearTuple(start=(0, 0), finish=(0, 0)):
    """
    Linearly increasing, but with two properites instead of one.
    """

    def linear_easing(sprite, delta):
        return ((finish[0] - start[0]) * delta + start[0],
                (finish[1] - start[1]) * delta + start[1])

    return linear_easing


def Arc(center=(0, 0), radius=1, theta_start=0, theta_end=2 * math.pi):
    """
    Increasing according to a circular curve for two properties.
    """

    def arc_easing(sprite, delta):
        theta = (theta_end - theta_start) * delta
        return (center[0] + radius * math.cos(theta),
                center[1] + radius * math.sin(theta))

    return arc_easing


def Polar(center=(0, 0),
          radius=lambda theta: 1.0,
          theta_start=0,
          theta_end=2 * math.pi):
    """
    Similar to an Arc, except the radius should be a function of time.
    """

    def arc_easing(sprite, delta):
        theta = (theta_end - theta_start) * delta
        return (center[0] + radius(theta) * math.cos(theta),
                center[1] + radius(theta) * math.sin(theta))

    return arc_easing


def Random(lowest=0, highest=100):
    """
    Random function between lowest and highest
    """

    def random_easing(sprite, delta):
        return random.randint(lowest, highest)

    return random_easing



import difflib
import importlib
import json
import os
import re
from os import path

from designer.utilities.argument_checks import make_suggestions

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'data/colors.json'), encoding='utf-8') as f:
    colors = json.load(f)
    f.close()


def hex_to_rgb(hx, hsl=False):
    """
    Converts a HEX code into RGB or HSL.
    All credit goes to here: https://stackoverflow.com/a/62083599/1718155
    CC BY-SA 4.0

    Args:
        hx (str): Takes both short as well as long HEX codes.
        hsl (bool): Converts the given HEX code into HSL value if True.
    Return:
        Tuple of length 3 consisting of either int or float values.
    Raise:
        ValueError: If given value is not a valid HEX code.
    """
    if re.compile(r'#[a-fA-F0-9]{3}(?:[a-fA-F0-9]{3})?$').match(hx):
        div = 255.0 if hsl else 0
        if len(hx) <= 4:
            return tuple(int(hx[i]*2, 16) / div if div else
                         int(hx[i]*2, 16) for i in (1, 2, 3))
        return tuple(int(hx[i:i+2], 16) / div if div else
                     int(hx[i:i+2], 16) for i in (1, 3, 5))
    raise ValueError(f'"{hx}" is not a valid HEX code. Should be something like "#FFB6C1".')


def _process_color(color) -> 'List[int]':
    """
    Converts either a color string (e.g., 'orange') into a color triplet (list of 3 numbers).
    If a color triplet is passed in, it is returned unmodified.

    :param color:
    :return:
    """
    if isinstance(color, str):
        if color and color[0] == '#':
            return hex_to_rgb(color)
        elif color not in colors:
            suggestions = make_suggestions(color, colors.keys(), cutoff=.3)
            if suggestions:
                raise ValueError(f"Unknown color {color!r}. Perhaps you meant one of these? {suggestions}")
            else:
                raise ValueError(f"Unknown color {color!r}. You should check the documentation!")
        return colors[color]
    if isinstance(color, int):
        raise ValueError(f"An integer ({color}) was given as the color value. The color should be a string ('red') or three numbers in a list.")
    else:
        return color



"""A vector is a class that behaves like a 2-tuple, but with convenient
methods."""

import math


class Vec2D:
    """
    Vec2D is a class that behaves like a 2-tuple, but with a number
    of convenient methods for vector calculation and manipulation.
    It can be created with two arguments for x,y, or by passing a
    2-tuple.

    In addition to the methods documented below, Vec2D supports
    the following:

    >>> from designer.utilities.vector import Vec2D
    >>> v1 = Vec2D(1,0)
    >>> v2 = Vec2D((0,1))    # Note 2-tuple argument!

    Tuple access, or x,y attribute access

    >>> v1.x
    1
    >>> v1.y
    0
    >>> v1[0]
    1
    >>> v1[1]
    0

    Addition, subtraction, and multiplication

    >>> v1 + v2
    Vec2D(1, 1)
    >>> v1 - v2
    Vec2D(1, -1)
    >>> 3 * v1
    Vec2D(3, 0)
    >>> (3, 4) * (v1+v2)
    Vec2D(3, 4)

    Compatibility with standard tuples

    >>> v1 + (1,1)
    Vec2D(2, 1)
    >>> (1,1) + v1
    Vec2D(2, 1)

    """
    __slots__ = ['x', 'y']

    def __init__(self, *args):
        if len(args) == 1:
            self.x, self.y = args[0]
        elif len(args) == 2:
            self.x, self.y = args[0], args[1]
        else:
            raise ValueError("Invalid Vec2D arguments")

    def __len__(self):
        return 2

    def __getitem__(self, key):
        if key == 0:
            return self.x
        if key == 1:
            return self.y
        raise IndexError("Invalid subscript %s" % (str(key)))

    def __repr__(self):
        return '(%s, %s)' % (str(self.x), str(self.y))

    def __eq__(self, o):
        try:
            return self.x == o[0] and self.y == o[1]
        except (IndexError, TypeError):
            pass
        if isinstance(o, (int, float)):
            return self.x == o and self.y == o
        return False

    def __ne__(self, o):
        return not self.__eq__(o)

    def __add__(self, o):
        try:
            return Vec2D(self.x + o[0], self.y + o[1])
        except (IndexError, TypeError):
            pass
        if isinstance(o, (int, float)):
            return Vec2D(self.x + o, self.y + o)
        return NotImplemented

    __radd__ = __add__

    def __sub__(self, o):
        try:
            return Vec2D(self.x - o[0], self.y - o[1])
        except (IndexError, TypeError):
            pass
        if isinstance(o, (int, float)):
            return Vec2D(self.x - o, self.y - o)
        return NotImplemented

    __isub__ = __sub__

    def __rsub__(self, o):
        try:
            return Vec2D(o[0] - self.x, o[1] - self.y)
        except (IndexError, TypeError):
            pass
        if isinstance(o, (int, float)):
            return Vec2D(o - self.x, o - self.y)
        return NotImplemented

    def __mul__(self, o):
        try:
            return Vec2D(self.x * o[0], self.y * o[1])
        except (IndexError, TypeError):
            pass

        if isinstance(o, (int, float)):
            return Vec2D(self.x * o, self.y * o)

        return NotImplemented

    __rmul__ = __mul__
    __imul__ = __mul__

    def __div__(self, o):
        try:
            return Vec2D(self.x / o[0], self.y / o[1])
        except (IndexError, TypeError):
            pass

        if isinstance(o, (int, float)):
            return Vec2D(self.x / o, self.y / o)

        return NotImplemented()

    __truediv__ = __div__

    def __neg__(self):
        return -Vec2D(self.x, -self.y)

    def __pos__(self):
        return self

    def get_length(self):
        """
        Return the length of this vector.

        :rtype: float
        """
        return math.sqrt(self.x * self.x + self.y * self.y)

    def get_length_squared(self):
        """
        Return the squared length of this vector.

        :rtype: int
        """
        return self.x * self.x + self.y * self.y

    def get_angle(self):
        """
        Return the angle this vector makes with the positive x axis.

        :rtype: float
        """
        return math.atan2(self.y, self.x)

    def perpendicular(self):
        """
        Returns a new :class:`Vec2D <designer.utilities.vector.Vec2D>` perpendicular to this one.

        :rtype: :class:`Vec2D <designer.utilities.vector.Vec2D>`
        """
        return Vec2D(-self.y, self.x)

    def dot(self, other):
        """
        Returns the `dot product <https://en.wikipedia.org/wiki/Dot_product>`_
        of this point with another.

        :param other: the other point
        :type other: 2-tuple or :class:`Vec2D <designer.utilities.vector.Vec2D>`
        :rtype: int
        """
        return self.x * other[0] + self.y * other[1]

    def distance(self, other):
        """
        Returns the distance from this :class:`Vec2D <designer.utilities.vector.Vec2D>` to the
        other point.

        :param other: the other point
        :type other: 2-tuple or :class:`Vec2D <designer.utilities.vector.Vec2D>`
        :rtype: float
        """
        return (other - self).get_length()

    def angle(self, other):
        """
        Returns the angle between this point and another point.

        :param other: the other point
        :type other: 2-tuple or :class:`Vec2D <designer.utilities.vector.Vec2D>`
        :rtype: float
        """
        x = self.x*other[1] - self.y*other[0]
        d = self.x*other[0] + self.y*other[1]
        return math.atan2(x, d)

    def projection(self, other):
        """
        Returns the
        `projection <https://en.wikipedia.org/wiki/Vector_projection>`_
        of this :class:`Vec2D <designer.utilities.vector.Vec2D>` onto another point.

        :param other: the other point
        :type other: 2-tuple or :class:`Vec2D <designer.utilities.vector.Vec2D>`
        :rtype: float
        """
        other = Vec2D(other)
        l2 = float(other.x*other.x + other.y*other.y)
        d = self.x*other.x + self.y*other.y
        return (d/l2)*other

    def rotated(self, angle, center=(0, 0)):
        """
        Returns a new vector from the old point rotated by `angle` radians about
        the optional `center`.

        :param angle: angle in radians.
        :type angle: float
        :param center: an optional center
        :type center: 2-tuple or :class:`Vec2D <designer.utilities.vector.Vec2D>`
        :rtype: :class:`Vec2D <designer.utilities.vector.Vec2D>`
        """
        p = self - center
        c = math.cos(angle)
        s = math.sin(angle)
        x = p.x*c - p.y*s
        y = p.x*s + p.y*c
        return Vec2D(x, y) + center

    def normalized(self):
        """
        Returns a new vector based on this one, normalized to length 1. That is,
        it keeps the same angle, but its length is now 1.

        :rtype: :class:`Vec2D <designer.utilities.vector.Vec2D>`
        """
        length = self.get_length()
        if length == 0:
            return None
        return Vec2D(self.x/length, self.y/length)

    def floor(self):
        """
        Converts the components of this vector into ints, discarding anything
        past the decimal place.

        :returns: this :class:`Vec2D <designer.utilities.vector.Vec2D>`
        """
        self.x = int(self.x)
        self.y = int(self.y)
        return self

    def to_polar(self):
        """
        Returns `Vec2D(radius, theta)` for this vector, where `radius` is the
        length and `theta` is the angle.

        :rtype: :class:`Vec2D <designer.utilities.vector.Vec2D>`
        """
        return Vec2D(self.get_length(), self.get_angle())

    @staticmethod
    def from_polar(*args):
        """
        Takes in radius, theta or (radius, theta) and returns rectangular
        :class:`Vec2D <designer.utilities.vector.Vec2D>`.

        :rtype: :class:`Vec2D <designer.utilities.vector.Vec2D>`
        """
        v = Vec2D(*args)
        return Vec2D(v.x*math.cos(v.y), v.x*math.sin(v.y))

    def __hash__(self):
        return hash(self.x + self.y)



"""
The LayerTree class manages the layers for a Scene; this is a complicated
problem, because we need to know the depth order for sprites extremely quickly
when we draw. All layers are precalculated by converting Relative Depth Chains
into an Absolute Position Value.

Important concepts:
    Relative Depth Chain - a list of numbers indicating the current relative
    position of this layer. If we had four layers like so:
        Scene -> View : [0,0]
        Scene -> View -> "top" : [0, 0, 0]
        Scene -> View -> "top" -> View : [0, 0, 1, 0]
        Scene -> View -> "bottom" : [0, 0, 1]
    Absolute Position Value
        A number representing a Relative Depth Chain collapsed into a single
        integer (or long, possibly)
"""

from weakref import ref as _wref

class _LayerTree:
    """
    Starts keeping track of the entity as a child of this view.

    :param scene: The scene that owns this LayerTree.
    :type scene: Scene (not a weakref).
    """
    #: The maximum number of layers for any given view/scene. After this, depth
    #: calculations will be messed up. It is an artificially chosen number, it
    #: should eventually be possible to change it.
    MAX_LAYERS = 40
    def __init__(self, scene):
        self.layers = {_wref(scene) : []}
        self.child_views = {_wref(scene) : []}
        self.layer_location = {_wref(scene) : [0]}
        self.scene = _wref(scene)
        self.tree_height = {_wref(scene) : 1}
        self._precompute_positions()
        self.maximum_height = 1

    def remove_view(self, view):
        """
        Removes all references to this view; it must have previously been added
        to the LayerTree.

        :param view: the View to remove
        :type view: View (not a weakref)
        """
        view = _wref(view)
        del self.tree_height[view]
        del self.layers[view]
        self.child_views[view()._parent].remove(view)
        del self.child_views[view]
        self._precompute_positions()

    def add_view(self, view):
        """
        Starts keeping track of this view in the LayerTree.

        :param view: the new View to add
        :type view: View (not a weakref)
        """
        parent = view._parent
        view = _wref(view)
        self.layers[view] = []
        self.child_views[view] = []
        self.child_views[parent].append(view)
        self.tree_height[view] = 1
        if len(self.child_views[parent]) == 1:
            self.tree_height[parent] += 1
            while parent != self.scene:
                parent = parent()._parent
                self.tree_height[parent] += 1
        self._precompute_positions()

    def set_view_layer(self, view, layer):
        """
        Set the layer that this View is on. Behavior is undefined if that layer
        does not exist in the parent, so make sure you eventually add that layer
        to the parent.

        :param view: the view have its layer set
        :type view: View (not a weakref)
        :param layer: the name of the layer on the parent
        :type layer: string
        """
        view.layer = layer
        self._precompute_positions()

    def set_view_layers(self, view, layers):
        """
        Set the layers that will be available for this view or scene.

        :param view: the view have its layer set
        :type view: View (not a weakref) or a Scene (not a weakref)
        :param layers: the name of the layer on the parent
        :type layers: a list of strings
        """
        self.layers[_wref(view)] = list(layers)
        self._precompute_positions()

    def _compute_positional_chain(self, chain):
        """
        From a list of numbers indicating the location of the View/layer within
        the current level, compute an absolute number in base MAX_LAYERS that
        can quickly and easily compared.

        :param chain: The relative positions at this level
        :type chain: a list of numbers
        :returns: An `int` representing the absolute position depth.
        """
        total = 0
        for index, value in enumerate(chain):
            power = self.maximum_height - index - 1
            total += value * (self.MAX_LAYERS ** power)
        return total

    def _precompute_positions(self):
        """
        Runs through the entire LayerTree and calculates an absolute number for
        each possible view/layer, which can be easily compared.
        """
        self.maximum_height = self.tree_height[self.scene]
        self.layer_location = {}
        self._precompute_position_for_layer(self.scene, [])
        for layer_key, v in self.layer_location.items():
            self.layer_location[layer_key] = self._compute_positional_chain(v)

    def _precompute_position_for_layer(self, view, current_position):
        """
        For a given view, and the current depth in the layer heirarchy, compute
        what its relative depth chain should look like. Sets this entry for
        layer_location to be a list of numbers indicating the relative depth.
        Note that this is called in a recursive manner to move down the entire
        LayerTree.

        :param view: The current view to explore
        :type view: a weakref to a View!
        :param current_position: The current relative depth chain
        :type current_position: list of numbers
        """
        position = 0
        for position, layer in enumerate(self.layers[view], 1):
            self.layer_location[(view, layer)] = current_position + [position]
        self.layer_location[view] = current_position + [1+position]
        for subview in self.child_views[view]:
            if subview().layer is None:
                new_position = self.layer_location[view]
            else:
                new_position = self.layer_location[(view, subview().layer)]
            self._precompute_position_for_layer(subview, new_position)

    def get_layer_position(self, parent, layer):
        """
        For a given layer (and also that layer's parent/owner, since layer
        alone is ambiguous), identify what the Absolute Position Value is from
        the precomputed cache, allowing for :above and :below modifiers.

        :param parent: The view or scene that has this layer
        :type parent: a View or a Scene (not weakrefs!)
        :param layer: The name of the layer that we're interested in.
        :type layer: string
        :returns: A `float` representing where this layer is relative to others.
        """
        parent = _wref(parent)
        s = layer.split(':')
        layer = s[0]
        offset = 0
        if len(s) > 1:
            mod = s[1]
            if mod == 'above':
                offset = 0.5
            if mod == 'below':
                offset = -0.5
        if (parent, layer) in self.layer_location:
            position = self.layer_location[(parent, layer)]
        elif parent in self.layer_location:
            position = self.layer_location[parent]
        else:
            position = self.layer_location[self.scene]
        return position + offset



"""This module contains functions and classes for creating and issuing events.
For a list of the events that are built into Designer, check the
:ref:`Event List<ref.events>`.

    .. attribute:: keys

        A special attribute for accessing the constants associated with a given
        key. For instance, ``designer.keys.down`` and ``designer.keys.f``. This is
        useful for testing for keyboard events. A complete list of all the key
        constants can be found in the
        :ref:`Keyboard Keys <ref.keys>` appendix.

    .. attribute:: mods

        A special attribute for accessing the constants associated with a given
        mod key. For instance, ``designer.mods.lshift`` (left shift) and
        ``designer.mods.ralt`` (Right alt). This is useful for testing for keyboard
        events. A complete list of all the key
        constants can be found in the
        :ref:`Keyboard Modifiers <ref.mods>` appendix.

"""
import inspect

import pygame

import json
import os
import random
import base64

import designer
from weakref import WeakMethod
from designer.utilities.weak_method import WeakMethod as _wm


def WeakMethod(func):
    try:
        return _wm(func)
    except TypeError:
        return func


class GameEndException(Exception):
    pass


class Event:
    """
    A class for building for attaching data to an event.
    Keyword arguments will be named attributes of the Event when it is passed
    into :func:`queue <designer.core.event.queue>`::

        collision_event = Event(ball=ball, paddle=paddle)
        designer.core.event.queue("ball.collides.paddle", collision_event)
    """

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __repr__(self):
        return f"<Event {self.__dict__!r}>"


# This might actually be unused!
_EVENT_NAMES = ['QUIT', 'ACTIVEEVENT', 'USEREVENT',
                'KEYDOWN', 'KEYUP',
                'MOUSEMOTION', 'MOUSEBUTTONUP', 'MOUSEBUTTONDOWN',
                'VIDEORESIZE', 'VIDEOEXPOSE',
                ]
MOUSE_MAP = ['left', 'middle', 'right', 'scroll_up', 'scroll_down']

_TYPE_TO_ATTRS = {
    pygame.QUIT: tuple(),
    pygame.ACTIVEEVENT: ('gain', 'state'),
    pygame.KEYDOWN: ('key', 'unicode', 'mod'),
    pygame.KEYUP: ('key', 'mod'),
    pygame.MOUSEMOTION: ('pos', 'rel', 'buttons'),
    pygame.MOUSEBUTTONUP: ('pos', 'button'),
    pygame.MOUSEBUTTONDOWN: ('pos', 'button'),
    pygame.VIDEORESIZE: ('size', 'w', 'h'),
    pygame.VIDEOEXPOSE: tuple(),
    pygame.AUDIODEVICEADDED: ('which', 'iscapture'),
    pygame.AUDIODEVICEREMOVED: ('which', 'iscapture'),
    pygame.FINGERMOTION: ('touch_id', 'finger_id', 'x', 'y', 'dx', 'dy'),
    pygame.FINGERDOWN: ('touch_id', 'finger_id', 'x', 'y', 'dx', 'dy'),
    pygame.FINGERUP: ('touch_id', 'finger_id', 'x', 'y', 'dx', 'dy'),
    pygame.MOUSEWHEEL: ('which', 'flipped', 'x', 'y'),
    pygame.MULTIGESTURE: ('touch_id', 'x', 'y', 'pinched', 'rotated', 'num_fingers'),
    pygame.TEXTEDITING: ('text', 'start', 'length'),
    pygame.TEXTINPUT: ('text', ),
    #pygame.WINDOWEVENT: ('event', ),
}

COMMON_EVENT_NAMES = {
    'drawing': 'director.render',
    'starting': 'director.start',
    'updating': 'director.update',
    'quitting': 'system.quit',
    'typing': 'input.keyboard.down',
    'done typing': 'input.keyboard.up',
    'scrolling': 'input.mouse.wheel',
    'tapping': 'input.touch.gesture',
    'clicking': 'input.mouse.down',
    'done clicking': 'input.mouse.down',
    'mouse motion': 'input.mouse.motion',
}

COMMON_EVENT_NAME_LOOKUP = {v: k for k, v in COMMON_EVENT_NAMES.items()}

_TYPE_TO_TYPE = {
    pygame.QUIT: "system.quit",
    pygame.ACTIVEEVENT: "system.focus_change",
    pygame.KEYDOWN: "input.keyboard.down",
    pygame.KEYUP: "input.keyboard.up",
    pygame.MOUSEMOTION: "input.mouse.motion",
    pygame.MOUSEBUTTONUP: "input.mouse.up",
    pygame.MOUSEBUTTONDOWN: "input.mouse.down",
    pygame.VIDEORESIZE: "system.video_resize",
    pygame.VIDEOEXPOSE: "system.video_expose",
    pygame.AUDIODEVICEADDED: 'system.audio_added',
    pygame.AUDIODEVICEREMOVED: 'system.audio_removed',
    pygame.FINGERMOTION: 'input.finger.motion',
    pygame.FINGERDOWN: 'input.finger.down',
    pygame.FINGERUP: 'input.finger.up',
    pygame.MOUSEWHEEL: 'input.mouse.wheel',
    pygame.MULTIGESTURE: 'input.touch.gesture',
    pygame.TEXTEDITING: 'input.text.editing',
    pygame.TEXTINPUT: 'input.text.input',
}

KNOWN_EVENTS = [*_TYPE_TO_TYPE.values(), *COMMON_EVENT_NAMES.keys()]


def get_positional_event_parameters(event_type: str, event):
    if event_type in ("updating", "director.update"):
        return "world", "delta"
    elif event_type in ("typing", ) or event_type.startswith("input.keyboard"):
        return "world", "key", "modifier", "character"
    elif event_type.startswith("input.mouse.motion"):
        return "world", "x", "y", "left", "middle", "right"
    elif event_type in ("clicking",) or event_type.startswith("input.mouse"):
        return "world", "x", "y", "button"
    elif event_type in ("starting", "director.start"):
        return "window"
    return [key for key in dir(event) if not key.startswith("__")]


def queue(event_name, event=None):
    """
    Queues a new event in the system, meaning that it will be run at the next
    available opportunity.

    :param str event_name: The type of event (e.g., ``"system.quit"``,
                           ``"input.mouse.up"``, or ``"pong.score"``.
    :param event: An Event object that holds properties for the event.
    :type event: :class:`Event <spyral.event.Event>`
    """
    designer.GLOBAL_DIRECTOR.current_window._queue_event(event_name, event)


def handle(event_name, event=None):
    """
    Instructs spyral to execute the handlers for this event right now. When you
    have a custom event, this is the function you call to have the event occur.

    :param str event_name: The type of event (e.g., ``"system.quit"``,
                           ``"input.mouse.up"``, or ``"pong.score"``.
    :param event: An Event object that holds properties for the event.
    :type event: :class:`Event <spyral.event.Event>`
    """
    designer.GLOBAL_DIRECTOR.current_window._handle_event(event_name, event)


def register(event_namespace, handler,
             args=None, kwargs=None, priority=0):
    """
    Registers an event `handler` to a namespace. Whenever an event in that
    `event_namespace` is fired, the event `handler` will execute with that
    event.

    :param event_namespace: the namespace of the event, e.g.
                            ``"input.mouse.left.click"`` or ``"pong.score"``.
    :type event_namespace: str
    :param handler: A function that will handle the event. The first
                    argument to the function will be the event.
    :type handler: function
    :param args: any additional arguments that need to be passed in
                 to the handler.
    :type args: sequence
    :param kwargs: any additional keyword arguments that need to be
                   passed into the handler.
    :type kwargs: dict
    :param int priority: the higher the `priority`, the sooner this handler will
                         be called in reaction to the event, relative to the
                         other event handlers registered.
    """
    event_namespace = COMMON_EVENT_NAMES.get(event_namespace, event_namespace)
    designer.GLOBAL_DIRECTOR.current_window._reg_internal(event_namespace, (WeakMethod(handler),),
                                                          args, kwargs, priority, False)


def unregister(event_namespace, handler):
    """
    Unregisters a registered handler for that namespace. Dynamic handler
    strings are supported as well.

    :param str event_namespace: An event namespace
    :param handler: The handler to unregister.
    :type handler: a function or string.
    """
    designer.GLOBAL_DIRECTOR.current_window._unregister(event_namespace, handler)


def clear_namespace(namespace):
    """
    Clears all handlers from namespaces that are at least as specific as the
    provided `namespace`.

    :param str namespace: The complete namespace.
    """
    designer.GLOBAL_DIRECTOR.current_window._clear_namespace(namespace)


class KeyboardKey:
    def __init__(self, value):
        self._value = value

    def __eq__(self, other):
        if isinstance(other, KeyboardKey):
            return self._value == other._value
        elif isinstance(other, int):
            return self._value == other
        elif isinstance(other, str):
            return self._value == pygame.key.key_code(other)
        else:
            raise TypeError(f"You attempted to compare a string Key ({self!r}) with the value {other!r}, but those types were not compatible.")

    def __repr__(self):
        return pygame.key.name(self._value)


def _pygame_to_spyral(event, **kwargs):
    """
    Convert a Pygame event to a Spyral event, correctly converting arguments to
    attributes.
    """
    event_attrs = _TYPE_TO_ATTRS.get(event.type, tuple())
    event_type = _TYPE_TO_TYPE.get(event.type, f'unknown.event[{event.type}]')
    e = Event(**kwargs)
    for attr in event_attrs:
        setattr(e, attr, getattr(event, attr))
    if event_type.startswith("input"):
        setattr(e, "type", event_type.split(".")[-1])
    if event_type.startswith('input.keyboard'):
        k = keys.reverse_map.get(event.key, 'unknown')
        event_type += '.' + k
        e.key = KeyboardKey(event.key)
        e.modifier = event.mod
        e.character = event.unicode
    if event_type.startswith('input.mouse.motion'):
        e.left, e.middle, e.right = map(bool, event.buttons)
    elif event_type.startswith('input.mouse.up') or event_type.startswith('input.mouse.down'):
        try:
            m = MOUSE_MAP[event.button - 1]
            setattr(e, "button", m)
        except IndexError:
            m = str(event.button)
        event_type += '.' + m
    if event_type.startswith('input.mouse.wheel'):
        e.pos = designer.utilities.Vec2D(e.x, e.y) / designer.GLOBAL_DIRECTOR.current_window._scale
        e.x = e.pos[0]
        e.y = e.pos[1]
    elif event_type.startswith('input.mouse') or event_type.startswith('input.touch'):
        e.pos = designer.utilities.Vec2D(e.pos) / designer.GLOBAL_DIRECTOR.current_window._scale
        e.x = e.pos[0]
        e.y = e.pos[1]

    return (event_type, e)


class EventHandler(object):
    """
    Base event handler class.
    """

    def __init__(self):
        self._events = []
        self._mouse_pos = (0, 0)

    def tick(self):
        """
        Should be called at the beginning of update cycle. For the
        event handler which is part of a scene, this function will be
        called automatically. For any additional event handlers, you
        must call this function manually.
        """
        pass

    def get(self, types=None):
        """
        Gets events from the event handler. Types is an optional
        iterable which has types which you would like to get.
        """
        if types is None:
            types = []
        try:
            types[0]
        except IndexError:
            pass
        except TypeError:
            types = (types,)

        if types == []:
            ret = self._events
            self._events = []
            return ret

        ret = [e for e in self._events if e['type'] in types]
        self._events = [e for e in self._events if e['type'] not in types]
        return ret


class LiveEventHandler(EventHandler):
    """
    An event handler which pulls events from the operating system.

    The optional output_file argument specifies the path to a file
    where the event handler will save a custom json file that can
    be used with the `ReplayEventHandler` to show replays of a
    game in action, or be used for other clever purposes.

    .. note::

        If you use the output_file parameter, this function will
        reseed the random number generator, save the seed used. It
        will then be restored by the ReplayEventHandler.
    """

    def __init__(self, output_file=None):
        EventHandler.__init__(self)
        self._save = output_file is not None
        if self._save:
            self._file = open(output_file, 'w')
            seed = os.urandom(4)
            info = {'random_seed': base64.encodestring(seed)}
            random.seed(seed)
            self._file.write(json.dumps(info) + "\n")

    def tick(self):
        mouse = pygame.mouse.get_pos()
        events = pygame.event.get()
        self._mouse_pos = mouse
        self._events.extend(events)
        # if self._save:
        #     d = {'mouse': mouse, 'events': events}
        #     self._file.write(json.dumps(d) + "\n")

    def __del__(self):
        if self._save:
            self._file.close()


class ReplayEventHandler(EventHandler):
    """
    An event handler which replays the events from a custom json
    file saved by the `LiveEventHandler`.
    """

    def __init__(self, input_file):
        EventHandler.__init__(self)
        self._file = open(input_file)
        info = json.loads(self._file.readline())
        random.seed(base64.decodestring(info['random_seed']))
        self.paused = False

    def pause(self):
        """
        Pauses the replay of the events, making tick() a noop until
        resume is called.
        """
        self.paused = True

    def resume(self):
        """
        Resumes the replay of events.
        """
        self.paused = False

    def tick(self):
        if self.paused:
            return
        try:
            d = json.loads(self._file.readline())
        except ValueError:
            # spyral.director.pop()
            pass
        events = d['events']
        events = [Event(e) for e in events]
        self._mouse_pos = d['mouse']
        self._events.extend(events)


class Mods(object):
    def __init__(self):
        self.none = pygame.KMOD_NONE
        self.lshift = pygame.KMOD_LSHIFT
        self.rshift = pygame.KMOD_RSHIFT
        self.shift = pygame.KMOD_SHIFT
        self.caps = pygame.KMOD_CAPS
        self.ctrl = pygame.KMOD_CTRL
        self.lctrl = pygame.KMOD_LCTRL
        self.rctrl = pygame.KMOD_RCTRL
        self.lalt = pygame.KMOD_LALT
        self.ralt = pygame.KMOD_RALT
        self.alt = pygame.KMOD_ALT


class Keys(object):

    def __init__(self):
        self.reverse_map = {}
        self.load_keys_from_file(designer.utilities.file_system.get_resource('default_key_mappings.txt'))
        self._fix_bad_names([("return", "enter"), ("break", "brk")])

    def _fix_bad_names(self, renames):
        """
        Used to replace any binding names with non-python keywords.
        """
        for original, new in renames:
            setattr(self, new, getattr(self, original))
            delattr(self, original)

    def load_keys_from_file(self, filename):
        fp = open(filename)
        key_maps = fp.readlines()
        fp.close()
        for single_mapping in key_maps:
            mapping = single_mapping[:-1].split(' ')
            if len(mapping) == 2:
                if mapping[1][0:2] == '0x':
                    setattr(self, mapping[0], int(mapping[1], 16))
                    self.reverse_map[int(mapping[1], 16)] = mapping[0]
                else:
                    setattr(self, mapping[0], int(mapping[1]))
                    self.reverse_map[int(mapping[1])] = mapping[0]

    def add_key_mapping(self, name, number):
        setattr(self, name, number)


keys = Keys()
mods = Mods()



import difflib
from designer.utilities.vector import Vec2D


def is_iterable_of_points(value):
    try:
        len(value)
        return all(is_point(v) for v in value)
    except:
        return False


def is_non_empty(value):
    try:
        value[0]
        return True
    except:
        return False

def at_least_three(value):
    try:
        return len(value) >= 3
    except:
        return False

def is_non_empty_iterable_of_points(value):
    return is_iterable_of_points(value) and is_non_empty(value)


def is_point(value):
    if isinstance(value, dict):
        return 'x' in value and 'y' in value and are_numbers(value['x'], value['y'])
    elif isinstance(value, (tuple, list)):
        return len(value) == 2 and are_numbers(value[0], value[1])
    elif isinstance(value, Vec2D):
        return True
    else:
        return hasattr(value, 'x') and hasattr(value, 'y') and are_numbers(value.x, value.y)


def are_numbers(*values):
    return all(isinstance(v, (int, float)) for v in values)


def make_suggestions(value, known_values, cutoff=.6):
    return ", ".join(map(repr, difflib.get_close_matches(value, known_values, cutoff=cutoff)))



import pygame

import math

from designer.colors import _process_color
from designer.helpers import get_width, get_height
from designer.objects.designer_object import DesignerObject
from designer.core.internal_image import InternalImage, DesignerSurface
from designer.utilities.vector import Vec2D
from designer.utilities.util import _anchor_offset


class Line(DesignerObject):
    FIELDS = (*DesignerObject.FIELDS,
              'start_x', 'start_y', 'end_x', 'end_y',
              'start', 'end', 'thickness', 'color')


    def __init__(self, start, end, thickness, color):
        """
        Creates Line Designer Object on window.

        :param start: starting coordinates of line
        :type start: Tuple[int]
        :param end: ending coordinates of line
        :type end: Tuple[int]
        :param thickness: thickness of line in pixels
        :type thickness: int
        :param color: color of line
        :type color: str or List[str]
        """
        super().__init__()
        self._color = color
        self._thickness = thickness
        self._calculate_positions(start, end, thickness)

        # And draw!
        self._redraw_internal_image()

    def _calculate_positions(self, start, end, thickness):
        (x1, y1), (x2, y2) = start, end
        # Need to flip y-axis
        angle = math.atan2(-(y2 - y1), x2 - x1) % 2 * math.pi
        p1x, p1y = x1 + math.cos(angle + math.pi / 2) * thickness, y1 + math.sin(angle + math.pi / 2) * thickness
        p2x, p2y = x1 + math.cos(angle - math.pi / 2) * thickness, y1 - math.sin(angle + math.pi / 2) * thickness
        p3x, p3y = x2 + math.cos(angle + math.pi / 2) * thickness, y2 + math.sin(angle + math.pi / 2) * thickness
        p4x, p4y = x2 + math.cos(angle - math.pi / 2) * thickness, y2 - math.sin(angle + math.pi / 2) * thickness
        left = min(p1x, p2x, p3x, p4x)
        right = max(p1x, p2x, p3x, p4x)
        top = min(p1y, p2y, p3y, p4y)
        bottom = max(p1y, p2y, p3y, p4y)

        # calculate differences between x and y coordinates of line
        x = abs(start[0] - end[0])
        y = abs(start[1] - end[1])

        width = right - left
        height = bottom - top

        new_start = (x1 - left), (y1 - top)
        new_end = (x2 - left), (y2 - top)

        self._pos = (left, top)
        self._start = new_start
        self._end = new_end
        self._size = (width, height)

    def _redraw_internal_image(self):
        # Scaling
        width = self._size[0] * self._scale[0]
        height = self._size[1] * self._scale[1]
        size = (int(width), int(height))
        color = _process_color(self._color)
        if size[0] <= 0 or size[1] <= 0:
            target = InternalImage(size=(1, 1)).fill(color)
            self._transform_image = target._surf
            self._recalculate_offset()
            self._expire_static()
            return
        new_image = InternalImage(size=size)
        new_image.draw_lines(color, [self._start, self._end], self._thickness, False)
        # Flip
        if self._flip_x or self._flip_y:
            new_image.flip(self._flip_x, self._flip_y)
        # Rotation
        if self._angle != 0:
            angle = self._angle % 360
            old = Vec2D(new_image.rect.center)
            new_image.rotate(angle)
            new = new_image.rect.center
            self._transform_offset = old - new
        self._transform_image = new_image._surf
        self._recalculate_offset()
        self._expire_static()

    @property
    def start_x(self):
        return self._start[0]

    @start_x.setter
    def start_x(self, value):
        self._start = (self._start[0], value)
        self._calculate_positions(self._start, self._end, self._thickness)
        self._redraw_internal_image()

    @property
    def start_y(self):
        return self._start[1]

    @start_y.setter
    def start_y(self, value):
        self._start = (value, self._start[1])
        self._calculate_positions(self._start, self._end, self._thickness)
        self._redraw_internal_image()

    @property
    def end_x(self):
        return self._end[0]

    @end_x.setter
    def end_x(self, value):
        self._end = (self._end[0], value)
        self._calculate_positions(self._start, self._end, self._thickness)
        self._redraw_internal_image()

    @property
    def end_y(self):
        return self._end[1]

    @end_y.setter
    def end_y(self, value):
        self._end = (value, self._end[1])
        self._calculate_positions(self._start, self._end, self._thickness)
        self._redraw_internal_image()

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value
        self._redraw_internal_image()

    @property
    def thickness(self):
        return self._thickness

    @thickness.setter
    def thickness(self, value):
        self._thickness = value
        self._redraw_internal_image()

def line(color, start_x, start_y, end_x=None, end_y=None, thickness=1):
    '''
    Function to create a line.

    :param thickness: thickness of line in pixels
    :type thickness: int
    :param color: color of line
    :type color: str or List[str]
    :param args: start and end coordinates of line
    :type args: two tuples (start x, start y), (end x, end y) or four ints start x, start y, end x, end y

    :return: Line object created
    '''
    if end_x is None and end_y is None:
        end_x, end_y = start_y
        start_x, start_y = start_x
    return Line((start_x, start_y), (end_x, end_y), thickness, color)


import pygame
import io
import os
from urllib.request import urlopen, Request


class SfxModule:
    _USER_AGENT = "Designer Game Library for Python"

    def __init__(self):
        pass

    def play(self, path, volume=1.0):
        sfx = self._load_sfx(path)
        sfx.set_volume(volume)
        sfx.play()
        return sfx

    def _load_sfx(self, filename):
        try:
            path_strs = filename.split('/')
            fixed_paths = os.path.join(*path_strs)
            if os.path.exists(fixed_paths):
                return pygame.mixer.Sound(fixed_paths)
            else:
                raise FileNotFoundError(fixed_paths)
        except FileNotFoundError as err:
            try:
                req = Request(filename, headers={'User-Agent': self._USER_AGENT})
                with urlopen(req) as opened_sfx:
                    sfx_str = opened_sfx.read()
                    sfx_file = io.BytesIO(sfx_str)
                    return pygame.mixer.Sound(sfx_file)
            except:
                if filename.startswith('https://') or filename.startswith('http://'):
                    raise ValueError(f"Unexpected error while playing url: {filename!r}")
                raise ValueError(f"Unexpected error while playing sound from filename: {filename!r}")




import imghdr
from typing import Optional

import pygame
import math

from designer.colors import _process_color
from designer.helpers import get_width, get_height
from designer.objects.designer_object import DesignerObject
from designer.core.internal_image import InternalImage, DesignerSurface
from designer.utilities import Vec2D, Rect
from designer.utilities.util import _anchor_offset
from designer.objects.image import Image


class DesignerGroup(DesignerObject):
    '''
    class consisting of group of DesignerObjects together to allow
    collective functionality
    '''

    FIELDS = (*DesignerObject.FIELDS, )

    def __init__(self, objects):
        """
        Creates Image Designer Object on window

        :param center: x, y coordinates of center of circle
        :type center: Tuple[int]
        """
        super().__init__()

        boxes = [
            Rect(object._pos, object._transform_image.get_size())
            for object in objects
        ]
        if boxes:
            combined = boxes[0]
            for other in boxes[1:]:
                combined = combined.union(other)
            center = combined.center
            topleft = combined.topleft
            size = combined.size
        else:
            center = get_width(), get_height()
            size = 1, 1
            topleft = center

        self._pos = center
        self._anchor = 'center'
        self._size = size
        #: Internal field holding the original version of the image
        self._internal_image: Optional[InternalImage] = InternalImage(size=(size))
        self._internal_image_version: Optional[int] = None

        for object in objects:
            self._internal_image._surf.blit(object._transform_image, object.pos - topleft)
            self._internal_image_version = 1

        # And draw!
        self._redraw_internal_image()

    def _recalculate_offset(self):
        """
        Recalculates this designer object's offset based on its position, transform
        offset, anchor, its image, and the image's scaling.
        """
        if self._internal_image is None:
            return
        size = self._scale * self._internal_image.size
        offset = _anchor_offset(self._anchor, size[0], size[1])
        self._offset = Vec2D(offset) - self._transform_offset

    def _redraw_internal_image(self):
        """
        Calculates the transforms that need to be applied to this designer object's
        image. In order: flipping, scaling, and rotation.
        """
        if self._internal_image is None:
            return
        source = self._internal_image._surf
        # Flip
        if self._flip_x or self._flip_y:
            source = pygame.transform.flip(source, self._flip_x, self._flip_y)
        # Scale
        if self._scale != (1.0, 1.0):
            new_size = self._scale * self._internal_image.size
            new_size = (int(new_size[0]), int(new_size[1]))
            if 0 in new_size:
                return self._make_blank_surface()
            new_surf = DesignerSurface(new_size)
            source = pygame.transform.smoothscale(source, new_size, new_surf)
        # Rotate
        if self._angle != 0:
            angle = self._angle % 360
            old = Vec2D(source.get_rect().center)
            source = pygame.transform.rotate(source, angle).convert_alpha()
            new = source.get_rect().center
            self._transform_offset = old - new
        # Finish updates
        self._transform_image = source
        self._recalculate_offset()
        self._expire_static()


def flatten(objects):
    if isinstance(objects, (list, tuple)):
        result = []
        for r in objects:
            result.extend(flatten(r))
        return result
    else:
        return [objects]


def group(*objects):
    '''
    Function to group multiple objects together.

    :param objects: collection of objects to be grouped together for collective functionality
    :type objects: at least one DesignerObject
    :return: Created Designer Group object
    '''
    return DesignerGroup(flatten(objects))
